from .igann import IGANN
from .igann import ELM_Regressor
from .igann_interactive import IGANN_interactive

# from .igann_bagged import IGANN_Bagged
from .igannSklearn import IGANNClassifier, IGANNRegressor
